/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.api.FeatureColourI;
import jalview.api.FeatureSettingsControllerI;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.SequenceI;
import jalview.datamodel.features.FeatureMatcherI;
import jalview.datamodel.features.FeatureMatcherSet;
import jalview.datamodel.features.FeatureMatcherSetI;
import jalview.datamodel.ontology.OntologyI;
import jalview.gui.Help.HelpId;
import jalview.io.JalviewFileChooser;
import jalview.io.JalviewFileView;
import jalview.io.gff.SequenceOntologyFactory;
import jalview.schemabinding.version2.Filter;
import jalview.schemabinding.version2.JalviewUserColours;
import jalview.schemabinding.version2.MatcherSet;
import jalview.schemes.FeatureColour;
import jalview.util.MessageManager;
import jalview.util.Platform;
import jalview.viewmodel.seqfeatures.FeatureRendererModel.FeatureSettingsBean;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.help.HelpSetException;
import javax.swing.AbstractCellEditor;
import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JLayeredPane;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.RowFilter;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableRowSorter;

public class FeatureSettings extends JPanel
        implements FeatureSettingsControllerI
{
  private static final Font VERDANA_12 = new Font("Verdana", Font.PLAIN, 12);

  private static final String SEQUENCE_FEATURE_COLOURS = MessageManager
          .getString("label.sequence_feature_colours");

  /*
   * column indices of fields in Feature Settings table
   */
  static final int TYPE_COLUMN = 0;

  static final int COLOUR_COLUMN = 1;

  static final int FILTER_COLUMN = 2;

  static final int SHOW_COLUMN = 3;

  private static final int COLUMN_COUNT = 4;

  private static final int MIN_WIDTH = 400;

  private static final int MIN_HEIGHT = 400;

  final FeatureRenderer fr;

  public final AlignFrame af;

  /*
   * 'original' fields hold settings to restore on Cancel
   */
  Object[][] originalData;

  private float originalTransparency;

  private Map<String, FeatureMatcherSetI> originalFilters;

  final JInternalFrame frame;

  JScrollPane scrollPane = new JScrollPane();

  JTable table;

  JPanel groupPanel;

  JSlider transparency = new JSlider();

  /*
   * when true, constructor is still executing - so ignore UI events
   */
  protected volatile boolean inConstruction = true;

  int selectedRow = -1;

  boolean resettingTable = false;

  /*
   * true when Feature Settings are updating from feature renderer
   */
  private boolean handlingUpdate = false;

  /*
   * holds {featureCount, totalExtent} for each feature type
   */
  Map<String, float[]> typeWidth = null;

  /*
   * if true, 'child' feature types are not displayed
   */
  JCheckBox summaryView;

  /*
   * those feature types that do not have a parent feature type present
   * (as determined by an Ontology relationship)
   */
  List<String> topLevelTypes;

  /**
   * Constructor
   * 
   * @param af
   */
  public FeatureSettings(AlignFrame alignFrame)
  {
    this.af = alignFrame;
    fr = af.getFeatureRenderer();

    // save transparency for restore on Cancel
    originalTransparency = fr.getTransparency();
    int originalTransparencyAsPercent = (int) (originalTransparency * 100);
    transparency.setMaximum(100 - originalTransparencyAsPercent);

    originalFilters = new HashMap<>(fr.getFeatureFilters()); // shallow copy

    topLevelTypes = new ArrayList<>();

    try
    {
      jbInit();
    } catch (Exception ex)
    {
      ex.printStackTrace();
    }

    initTable();

    scrollPane.setViewportView(table);

    if (af.getViewport().isShowSequenceFeatures() || !fr.hasRenderOrder())
    {
      fr.findAllFeatures(true); // display everything!
    }

    discoverAllFeatureData();
    final PropertyChangeListener change;
    final FeatureSettings fs = this;
    fr.addPropertyChangeListener(change = new PropertyChangeListener()
    {
      @Override
      public void propertyChange(PropertyChangeEvent evt)
      {
        if (!fs.resettingTable && !fs.handlingUpdate)
        {
          fs.handlingUpdate = true;
          fs.resetTable(null);
          // new groups may be added with new sequence feature types only
          fs.handlingUpdate = false;
        }
      }
    });

    frame = new JInternalFrame();
    frame.setContentPane(this);
    if (Platform.isAMac())
    {
      Desktop.addInternalFrame(frame,
              MessageManager.getString("label.sequence_feature_settings"),
              600, 480);
    }
    else
    {
      Desktop.addInternalFrame(frame,
              MessageManager.getString("label.sequence_feature_settings"),
              600, 450);
    }
    frame.setMinimumSize(new Dimension(MIN_WIDTH, MIN_HEIGHT));

    frame.addInternalFrameListener(
            new javax.swing.event.InternalFrameAdapter()
            {
              @Override
              public void internalFrameClosed(
                      javax.swing.event.InternalFrameEvent evt)
              {
                fr.removePropertyChangeListener(change);
              };
            });
    frame.setLayer(JLayeredPane.PALETTE_LAYER);
    inConstruction = false;
  }

  /**
   * Constructs and configures the JTable which displays columns of data for
   * each feature type
   */
  protected void initTable()
  {
    table = new JTable()
    {
      @Override
      public String getToolTipText(MouseEvent e)
      {
        String tip = null;
        int column = table.columnAtPoint(e.getPoint());
        switch (column)
        {
        case TYPE_COLUMN:
          /*
           * drag to reorder not enabled in Summary View
           */
          tip = summaryView.isSelected()
                  ? MessageManager.getString(
                          "label.feature_settings_select_columns")
                  : JvSwingUtils.wrapTooltip(true, MessageManager
                  .getString("label.feature_settings_click_drag"));
          break;
        case FILTER_COLUMN:
          int row = table.rowAtPoint(e.getPoint());
          FeatureMatcherSet o = (FeatureMatcherSet) table.getValueAt(row,
                  column);
          tip = o.isEmpty()
                  ? MessageManager.getString("label.filters_tooltip")
                  : o.toString();
          break;
        default:
          break;
        }
        return tip;
      }
    };
    table.getTableHeader().setFont(VERDANA_12);
    table.setFont(VERDANA_12);

    table.setDefaultEditor(FeatureColour.class, new ColorEditor(this));
    table.setDefaultRenderer(FeatureColour.class, new ColorRenderer());

    table.setDefaultEditor(FeatureMatcherSet.class, new FilterEditor(this));
    table.setDefaultRenderer(FeatureMatcherSet.class, new FilterRenderer());

    TableColumn colourColumn = new TableColumn(COLOUR_COLUMN, 75,
            new ColorRenderer(), new ColorEditor(this));
    table.addColumn(colourColumn);

    TableColumn filterColumn = new TableColumn(FILTER_COLUMN, 75,
            new FilterRenderer(), new FilterEditor(this));
    table.addColumn(filterColumn);

    table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

    table.addMouseListener(new MouseAdapter()
    {
      @Override
      public void mousePressed(MouseEvent evt)
      {
        selectedRow = table.rowAtPoint(evt.getPoint());
        String type = (String) table.getValueAt(selectedRow, TYPE_COLUMN);
        if (evt.isPopupTrigger())
        {
          Object colour = table.getValueAt(selectedRow, COLOUR_COLUMN);
          popupMenu(selectedRow, type, colour, evt.getX(), evt.getY());
        }
        else if (evt.getClickCount() == 2)
        {
          boolean invertSelection = evt.isAltDown();
          boolean toggleSelection = Platform.isControlDown(evt);
          boolean extendSelection = evt.isShiftDown();
          String[] terms = getTermsInScope(type);
          fr.ap.alignFrame.avc.markColumnsContainingFeatures(
                  invertSelection, extendSelection, toggleSelection, terms);
        }
      }

      // isPopupTrigger fires on mouseReleased on Windows
      @Override
      public void mouseReleased(MouseEvent evt)
      {
        selectedRow = table.rowAtPoint(evt.getPoint());
        if (evt.isPopupTrigger())
        {
          String type = (String) table.getValueAt(selectedRow, TYPE_COLUMN);
          Object colour = table.getValueAt(selectedRow, COLOUR_COLUMN);
          popupMenu(selectedRow, type, colour, evt.getX(), evt.getY());
        }
      }
    });

    table.addMouseMotionListener(new MouseMotionAdapter()
    {
      @Override
      public void mouseDragged(MouseEvent evt)
      {
        int newRow = table.rowAtPoint(evt.getPoint());
        dragRow(newRow);
      }
    });
  }

  /**
   * Answers an array consisting of the given type, and also (if 'Summary View'
   * is selected), any child terms in the sequence ontology
   * 
   * @param type
   * @return
   */
  protected String[] getTermsInScope(String type)
  {
    if (!summaryView.isSelected())
    {
      return new String[] { type };
    }

    List<String> terms = new ArrayList<>();
    terms.add(type);

    OntologyI so = SequenceOntologyFactory.getInstance();

    Object[][] data = ((FeatureTableModel) table.getModel()).getData();
    for (Object[] row : data)
    {
      String type2 = (String) row[TYPE_COLUMN];
      if (!type2.equals(type) && so.isA(type2, type))
      {
        terms.add(type2);
      }
    }
    return terms.toArray(new String[terms.size()]);
  }

  protected void popupMenu(final int rowSelected, final String type,
          final Object typeCol, int x, int y)
  {
    final FeatureColourI featureColour = (FeatureColourI) typeCol;

    JPopupMenu men = new JPopupMenu(MessageManager
            .formatMessage("label.settings_for_param", new String[]
            { type }));
    JMenuItem scr = new JMenuItem(
            MessageManager.getString("label.sort_by_score"));
    men.add(scr);
    final FeatureSettings me = this;
    scr.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        String[] types = getTermsInScope(type);
        me.af.avc.sortAlignmentByFeatureScore(Arrays.asList(types));
      }
    });
    JMenuItem dens = new JMenuItem(
            MessageManager.getString("label.sort_by_density"));
    dens.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        String[] types = getTermsInScope(type);
        me.af.avc.sortAlignmentByFeatureDensity(Arrays.asList(types));
      }
    });
    men.add(dens);

    /*
     * variable colour options include colour by label, by score,
     * by selected attribute text, or attribute value
     */
    final JCheckBoxMenuItem mxcol = new JCheckBoxMenuItem(
            MessageManager.getString("label.variable_colour"));
    mxcol.setSelected(!featureColour.isSimpleColour());
    men.add(mxcol);
    mxcol.addActionListener(new ActionListener()
    {
      JColorChooser colorChooser;

      @Override
      public void actionPerformed(ActionEvent e)
      {
        if (e.getSource() == mxcol)
        {
          if (featureColour.isSimpleColour())
          {
            FeatureTypeSettings fc = new FeatureTypeSettings(me.fr, type);
            fc.addActionListener(this);
          }
          else
          {
            // bring up simple color chooser
            colorChooser = new JColorChooser();
            String title = MessageManager
                    .getString("label.select_colour");
            JDialog dialog = JColorChooser.createDialog(me,
                    title, true, // modal
                    colorChooser, this, // OK button handler
                    null); // no CANCEL button handler
            colorChooser.setColor(featureColour.getMaxColour());
            dialog.setVisible(true);
          }
        }
        else
        {
          if (e.getSource() instanceof FeatureTypeSettings)
          {
            /*
             * update after OK in feature colour dialog; the updated
             * colour will have already been set in the FeatureRenderer
             */
            FeatureColourI fci = fr.getFeatureColours().get(type);
            table.setValueAt(fci, rowSelected, 1);
            table.validate();
          }
          else
          {
            // probably the color chooser!
            table.setValueAt(new FeatureColour(colorChooser.getColor()),
                    rowSelected, 1);
            table.validate();
            me.updateFeatureRenderer(
                    ((FeatureTableModel) table.getModel()).getData(),
                    false);
          }
        }
      }
    });

    JMenuItem selCols = new JMenuItem(
            MessageManager.getString("label.select_columns_containing"));
    selCols.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent arg0)
      {
        String[] types = getTermsInScope(type);
        fr.ap.alignFrame.avc.markColumnsContainingFeatures(false, false,
                false, types);
      }
    });
    JMenuItem clearCols = new JMenuItem(MessageManager
            .getString("label.select_columns_not_containing"));
    clearCols.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent arg0)
      {
        String[] types = getTermsInScope(type);
        fr.ap.alignFrame.avc.markColumnsContainingFeatures(true, false,
                false, types);
      }
    });
    JMenuItem hideCols = new JMenuItem(
            MessageManager.getString("label.hide_columns_containing"));
    hideCols.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent arg0)
      {
        String[] types = getTermsInScope(type);
        fr.ap.alignFrame.hideFeatureColumns(true, types);
      }
    });
    JMenuItem hideOtherCols = new JMenuItem(
            MessageManager.getString("label.hide_columns_not_containing"));
    hideOtherCols.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent arg0)
      {
        String[] types = getTermsInScope(type);
        fr.ap.alignFrame.hideFeatureColumns(false, types);
      }
    });
    men.add(selCols);
    men.add(clearCols);
    men.add(hideCols);
    men.add(hideOtherCols);
    men.show(table, x, y);
  }

  @Override
  synchronized public void discoverAllFeatureData()
  {
    Set<String> allGroups = new HashSet<>();
    AlignmentI alignment = af.getViewport().getAlignment();

    for (int i = 0; i < alignment.getHeight(); i++)
    {
      SequenceI seq = alignment.getSequenceAt(i);
      for (String group : seq.getFeatures().getFeatureGroups(true))
      {
        if (group != null && !allGroups.contains(group))
        {
          allGroups.add(group);
          checkGroupState(group);
        }
      }
    }

    resetTable(null);

    validate();
  }

  /**
   * Synchronise gui group list and check visibility of group
   * 
   * @param group
   * @return true if group is visible
   */
  private boolean checkGroupState(String group)
  {
    boolean visible = fr.checkGroupVisibility(group, true);

    for (int g = 0; g < groupPanel.getComponentCount(); g++)
    {
      if (((JCheckBox) groupPanel.getComponent(g)).getText().equals(group))
      {
        ((JCheckBox) groupPanel.getComponent(g)).setSelected(visible);
        return visible;
      }
    }

    final String grp = group;
    final JCheckBox check = new JCheckBox(group, visible);
    check.setFont(new Font("Serif", Font.BOLD, 12));
    check.setToolTipText(group);
    check.addItemListener(new ItemListener()
    {
      @Override
      public void itemStateChanged(ItemEvent evt)
      {
        fr.setGroupVisibility(check.getText(), check.isSelected());
        resetTable(new String[] { grp });
        af.alignPanel.paintAlignment(true, true);
      }
    });
    groupPanel.add(check);
    return visible;
  }

  synchronized void resetTable(String[] groupChanged)
  {
    if (resettingTable)
    {
      return;
    }
    resettingTable = true;
    typeWidth = new Hashtable<>();
    // TODO: change avWidth calculation to 'per-sequence' average and use long
    // rather than float

    Set<String> displayableTypes = new HashSet<>();
    Set<String> foundGroups = new HashSet<>();

    /*
     * determine which feature types may be visible depending on 
     * which groups are selected, and recompute average width data
     */
    for (int i = 0; i < af.getViewport().getAlignment().getHeight(); i++)
    {

      SequenceI seq = af.getViewport().getAlignment().getSequenceAt(i);

      /*
       * get the sequence's groups for positional features
       * and keep track of which groups are visible
       */
      Set<String> groups = seq.getFeatures().getFeatureGroups(true);
      Set<String> visibleGroups = new HashSet<>();
      for (String group : groups)
      {
        if (group == null || checkGroupState(group))
        {
          visibleGroups.add(group);
        }
      }
      foundGroups.addAll(groups);

      /*
       * get distinct feature types for visible groups
       * record distinct visible types, and their count and total length
       */
      Set<String> types = seq.getFeatures().getFeatureTypesForGroups(true,
              visibleGroups.toArray(new String[visibleGroups.size()]));

      for (String type : types)
      {
        displayableTypes.add(type);
        float[] avWidth = typeWidth.get(type);
        if (avWidth == null)
        {
          avWidth = new float[2];
          typeWidth.put(type, avWidth);
        }
        // todo this could include features with a non-visible group
        // - do we greatly care?
        // todo should we include non-displayable features here, and only
        // update when features are added?
        avWidth[0] += seq.getFeatures().getFeatureCount(true, type);
        avWidth[1] += seq.getFeatures().getTotalFeatureLength(type);
      }
    }

    /*
     * enable 'Summary View' if some types are sub-types of others
     */
    Set<String> parents = SequenceOntologyFactory.getInstance()
            .getParentTerms(displayableTypes);
    summaryView.setEnabled(parents.size() < displayableTypes.size());

    Object[][] data = new Object[displayableTypes.size()][COLUMN_COUNT];
    int dataIndex = 0;

    if (fr.hasRenderOrder())
    {
      if (!handlingUpdate)
      {
        fr.findAllFeatures(groupChanged != null); // prod to update
        // colourschemes. but don't
        // affect display
        // First add the checks in the previous render order,
        // in case the window has been closed and reopened
      }
      List<String> frl = fr.getRenderOrder();
      for (int ro = frl.size() - 1; ro > -1; ro--)
      {
        String type = frl.get(ro);

        if (!displayableTypes.contains(type))
        {
          continue;
        }
        data[dataIndex][TYPE_COLUMN] = type;
        data[dataIndex][COLOUR_COLUMN] = fr.getFeatureStyle(type);
        FeatureMatcherSetI featureFilter = fr.getFeatureFilter(type);
        data[dataIndex][FILTER_COLUMN] = featureFilter == null
                ? new FeatureMatcherSet()
                : featureFilter;
        data[dataIndex][SHOW_COLUMN] = new Boolean(
                af.getViewport().getFeaturesDisplayed().isVisible(type));
        dataIndex++;
        displayableTypes.remove(type);
      }
    }

    /*
     * process any extra features belonging only to 
     * a group which was just selected
     */
    while (!displayableTypes.isEmpty())
    {
      String type = displayableTypes.iterator().next();
      data[dataIndex][TYPE_COLUMN] = type;

      data[dataIndex][COLOUR_COLUMN] = fr.getFeatureStyle(type);
      if (data[dataIndex][COLOUR_COLUMN] == null)
      {
        // "Colour has been updated in another view!!"
        fr.clearRenderOrder();
        return;
      }
      FeatureMatcherSetI featureFilter = fr.getFeatureFilter(type);
      data[dataIndex][FILTER_COLUMN] = featureFilter == null
              ? new FeatureMatcherSet()
              : featureFilter;
      data[dataIndex][SHOW_COLUMN] = new Boolean(true);
      dataIndex++;
      displayableTypes.remove(type);
    }

    if (originalData == null)
    {
      originalData = new Object[data.length][COLUMN_COUNT];
      for (int i = 0; i < data.length; i++)
      {
        System.arraycopy(data[i], 0, originalData[i], 0, COLUMN_COUNT);
      }
    }
    else
    {
      updateOriginalData(data);
    }

    /*
     * recreate the table model
     */
    FeatureTableModel dataModel = new FeatureTableModel(data);
    table.setModel(dataModel);

    /*
     * we want to be able to filter out rows for sub-types, but not to sort 
     * rows, so have to add a RowFilter to a disabled TableRowSorter (!)
     */
    final TableRowSorter<FeatureTableModel> sorter = new TableRowSorter<>(
            dataModel);
    for (int i = 0; i < table.getColumnCount(); i++)
    {
      sorter.setSortable(i, false);
    }

    /*
     * filter rows to only top-level Ontology types if requested
     */
    sorter.setRowFilter(new RowFilter<FeatureTableModel, Integer>()
    {
      @Override
      public boolean include(
              Entry<? extends FeatureTableModel, ? extends Integer> entry)
      {
        if (!summaryView.isSelected())
        {
          return true;
        }
        int row = entry.getIdentifier(); // this is model, not view, row number
        String featureType = (String) entry.getModel().getData()[row][TYPE_COLUMN];
        return parents.contains(featureType);
      }
    });
    table.setRowSorter(sorter);

    table.getColumnModel().getColumn(0).setPreferredWidth(200);

    groupPanel.setLayout(
            new GridLayout(fr.getFeatureGroupsSize() / 4 + 1, 4));
    pruneGroups(foundGroups);
    groupPanel.validate();

    updateFeatureRenderer(data, groupChanged != null);
    resettingTable = false;
  }

  /**
   * Updates 'originalData' (used for restore on Cancel) if we detect that changes
   * have been made outwith this dialog
   * <ul>
   * <li>a new feature type added (and made visible)</li>
   * <li>a feature colour changed (in the Amend Features dialog)</li>
   * </ul>
   * 
   * @param foundData
   */
  protected void updateOriginalData(Object[][] foundData)
  {
    // todo LinkedHashMap instead of Object[][] would be nice

    Object[][] currentData = ((FeatureTableModel) table.getModel())
            .getData();
    for (Object[] row : foundData)
    {
      String type = (String) row[TYPE_COLUMN];
      boolean found = false;
      for (Object[] current : currentData)
      {
        if (type.equals(current[TYPE_COLUMN]))
        {
          found = true;
          /*
           * currently dependent on object equality here;
           * really need an equals method on FeatureColour
           */
          if (!row[COLOUR_COLUMN].equals(current[COLOUR_COLUMN]))
          {
            /*
             * feature colour has changed externally - update originalData
             */
            for (Object[] original : originalData)
            {
              if (type.equals(original[TYPE_COLUMN]))
              {
                original[COLOUR_COLUMN] = row[COLOUR_COLUMN];
                break;
              }
            }
          }
          break;
        }
      }
      if (!found)
      {
        /*
         * new feature detected - add to original data (on top)
         */
        Object[][] newData = new Object[originalData.length
                + 1][COLUMN_COUNT];
        for (int i = 0; i < originalData.length; i++)
        {
          System.arraycopy(originalData[i], 0, newData[i + 1], 0,
                  COLUMN_COUNT);
        }
        newData[0] = row;
        originalData = newData;
      }
    }
  }

  /**
   * Remove from the groups panel any checkboxes for groups that are not in the
   * foundGroups set. This enables removing a group from the display when the last
   * feature in that group is deleted.
   * 
   * @param foundGroups
   */
  protected void pruneGroups(Set<String> foundGroups)
  {
    for (int g = 0; g < groupPanel.getComponentCount(); g++)
    {
      JCheckBox checkbox = (JCheckBox) groupPanel.getComponent(g);
      if (!foundGroups.contains(checkbox.getText()))
      {
        groupPanel.remove(checkbox);
      }
    }
  }

  /**
   * reorder data based on the featureRenderers global priority list.
   * 
   * @param data
   */
  private void ensureOrder(Object[][] data)
  {
    boolean sort = false;
    float[] order = new float[data.length];
    for (int i = 0; i < order.length; i++)
    {
      order[i] = fr.getOrder(data[i][0].toString());
      if (order[i] < 0)
      {
        order[i] = fr.setOrder(data[i][0].toString(), i / order.length);
      }
      if (i > 1)
      {
        sort = sort || order[i - 1] > order[i];
      }
    }
    if (sort)
    {
      jalview.util.QuickSort.sort(order, data);
    }
  }

  /**
   * Offers a file chooser dialog, and then loads the feature colours and
   * filters from file in XML format and unmarshals to Jalview feature settings
   */
  void load()
  {
    JalviewFileChooser chooser = new JalviewFileChooser("fc",
            SEQUENCE_FEATURE_COLOURS);
    chooser.setFileView(new JalviewFileView());
    chooser.setDialogTitle(
            MessageManager.getString("label.load_feature_colours"));
    chooser.setToolTipText(MessageManager.getString("action.load"));

    int value = chooser.showOpenDialog(this);

    if (value == JalviewFileChooser.APPROVE_OPTION)
    {
      File file = chooser.getSelectedFile();
      load(file);
    }
  }

  /**
   * Loads feature colours and filters from XML stored in the given file
   * 
   * @param file
   */
  void load(File file)
  {
    try
    {
      InputStreamReader in = new InputStreamReader(
              new FileInputStream(file), "UTF-8");

      JalviewUserColours jucs = JalviewUserColours.unmarshal(in);

      /*
       * load feature colours
       */
      for (int i = jucs.getColourCount() - 1; i >= 0; i--)
      {
        jalview.schemabinding.version2.Colour newcol = jucs.getColour(i);
        FeatureColourI colour = Jalview2XML.unmarshalColour(newcol);
        fr.setColour(newcol.getName(), colour);
        fr.setOrder(newcol.getName(), i / (float) jucs.getColourCount());
      }

      /*
       * load feature filters; loaded filters will replace any that are
       * currently defined, other defined filters are left unchanged 
       */
      for (int i = 0; i < jucs.getFilterCount(); i++)
      {
        jalview.schemabinding.version2.Filter filterModel = jucs
                .getFilter(i);
        String featureType = filterModel.getFeatureType();
        FeatureMatcherSetI filter = Jalview2XML.unmarshalFilter(featureType,
                filterModel.getMatcherSet());
        if (!filter.isEmpty())
        {
          fr.setFeatureFilter(featureType, filter);
        }
      }

      /*
       * update feature settings table
       */
      if (table != null)
      {
        resetTable(null);
        Object[][] data = ((FeatureTableModel) table.getModel())
                .getData();
        ensureOrder(data);
        updateFeatureRenderer(data, false);
        table.repaint();
      }
    } catch (Exception ex)
    {
      System.out.println("Error loading User Colour File\n" + ex);
    }
  }

  /**
   * Offers a file chooser dialog, and then saves the current feature colours
   * and any filters to the selected file in XML format
   */
  void save()
  {
    JalviewFileChooser chooser = new JalviewFileChooser("fc",
            SEQUENCE_FEATURE_COLOURS);
    chooser.setFileView(new JalviewFileView());
    chooser.setDialogTitle(
            MessageManager.getString("label.save_feature_colours"));
    chooser.setToolTipText(MessageManager.getString("action.save"));

    int value = chooser.showSaveDialog(this);

    if (value == JalviewFileChooser.APPROVE_OPTION)
    {
      save(chooser.getSelectedFile());
    }
  }

  /**
   * Saves feature colours and filters to the given file
   * 
   * @param file
   */
  void save(File file)
  {
    JalviewUserColours ucs = new JalviewUserColours();
    ucs.setSchemeName("Sequence Features");
    try
    {
      PrintWriter out = new PrintWriter(new OutputStreamWriter(
              new FileOutputStream(file), "UTF-8"));

      /*
       * sort feature types by colour order, from 0 (highest)
       * to 1 (lowest)
       */
      Set<String> fr_colours = fr.getAllFeatureColours();
      String[] sortedTypes = fr_colours
              .toArray(new String[fr_colours.size()]);
      Arrays.sort(sortedTypes, new Comparator<String>()
      {
        @Override
        public int compare(String type1, String type2)
        {
          return Float.compare(fr.getOrder(type1), fr.getOrder(type2));
        }
      });

      /*
       * save feature colours
       */
      for (String featureType : sortedTypes)
      {
        FeatureColourI fcol = fr.getFeatureStyle(featureType);
        jalview.schemabinding.version2.Colour col = Jalview2XML.marshalColour(
                featureType, fcol);
        ucs.addColour(col);
      }

      /*
       * save any feature filters
       */
      for (String featureType : sortedTypes)
      {
        FeatureMatcherSetI filter = fr.getFeatureFilter(featureType);
        if (filter != null && !filter.isEmpty())
        {
          Iterator<FeatureMatcherI> iterator = filter.getMatchers().iterator();
          FeatureMatcherI firstMatcher = iterator.next();
          MatcherSet ms = Jalview2XML.marshalFilter(firstMatcher, iterator,
                  filter.isAnded());
          Filter filterModel = new Filter();
          filterModel.setFeatureType(featureType);
          filterModel.setMatcherSet(ms);
          ucs.addFilter(filterModel);
        }
      }

      ucs.marshal(out);
      out.close();
    } catch (Exception ex)
    {
      ex.printStackTrace();
    }
  }

  public void invertSelection()
  {
    Object[][] data = ((FeatureTableModel) table.getModel()).getData();
    for (int i = 0; i < data.length; i++)
    {
      data[i][SHOW_COLUMN] = !(Boolean) data[i][SHOW_COLUMN];
    }
    updateFeatureRenderer(data, true);
    table.repaint();
  }

  public void orderByAvWidth()
  {
    if (table == null || table.getModel() == null)
    {
      return;
    }
    Object[][] data = ((FeatureTableModel) table.getModel()).getData();
    float[] width = new float[data.length];
    float[] awidth;
    float max = 0;

    for (int i = 0; i < data.length; i++)
    {
      awidth = typeWidth.get(data[i][TYPE_COLUMN]);
      if (awidth[0] > 0)
      {
        width[i] = awidth[1] / awidth[0];// *awidth[0]*awidth[2]; - better
        // weight - but have to make per
        // sequence, too (awidth[2])
        // if (width[i]==1) // hack to distinguish single width sequences.
      }
      else
      {
        width[i] = 0;
      }
      if (max < width[i])
      {
        max = width[i];
      }
    }
    boolean sort = false;
    for (int i = 0; i < width.length; i++)
    {
      // awidth = (float[]) typeWidth.get(data[i][0]);
      if (width[i] == 0)
      {
        width[i] = fr.getOrder(data[i][TYPE_COLUMN].toString());
        if (width[i] < 0)
        {
          width[i] = fr.setOrder(data[i][TYPE_COLUMN].toString(),
                  i / data.length);
        }
      }
      else
      {
        width[i] /= max; // normalize
        fr.setOrder(data[i][TYPE_COLUMN].toString(), width[i]); // store for later
      }
      if (i > 0)
      {
        sort = sort || width[i - 1] > width[i];
      }
    }
    if (sort)
    {
      jalview.util.QuickSort.sort(width, data);
      // update global priority order
    }

    updateFeatureRenderer(data, false);
    table.repaint();
  }

  public void close()
  {
    try
    {
      frame.setClosed(true);
    } catch (Exception exe)
    {
    }

  }

  public void updateFeatureRenderer(Object[][] data)
  {
    updateFeatureRenderer(data, true);
  }

  /**
   * Update the priority order of features; only repaint if this changed the order
   * of visible features
   * 
   * @param data
   * @param visibleNew
   */
  private void updateFeatureRenderer(Object[][] data, boolean visibleNew)
  {
    FeatureSettingsBean[] rowData = getTableAsBeans(data);

    if (fr.setFeaturePriority(rowData, visibleNew))
    {
      af.alignPanel.paintAlignment(true, true);
    }
  }

  /**
   * Converts table data into an array of data beans
   */
  private FeatureSettingsBean[] getTableAsBeans(Object[][] data)
  {
    FeatureSettingsBean[] rowData = new FeatureSettingsBean[data.length];
    for (int i = 0; i < data.length; i++)
    {
      String type = (String) data[i][TYPE_COLUMN];
      FeatureColourI colour = (FeatureColourI) data[i][COLOUR_COLUMN];
      FeatureMatcherSetI theFilter = (FeatureMatcherSetI) data[i][FILTER_COLUMN];
      Boolean isShown = (Boolean) data[i][SHOW_COLUMN];
      rowData[i] = new FeatureSettingsBean(type, colour, theFilter,
              isShown);
    }
    return rowData;
  }

  private void jbInit() throws Exception
  {
    this.setLayout(new BorderLayout());

    JPanel settingsPane = new JPanel();
    settingsPane.setLayout(new BorderLayout());

    JPanel bigPanel = new JPanel();
    bigPanel.setLayout(new BorderLayout());

    groupPanel = new JPanel();
    bigPanel.add(groupPanel, BorderLayout.NORTH);

    JButton invert = new JButton(
            MessageManager.getString("label.invert_selection"));
    invert.setFont(JvSwingUtils.getLabelFont());
    invert.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        invertSelection();
      }
    });

    JButton optimizeOrder = new JButton(
            MessageManager.getString("label.optimise_order"));
    optimizeOrder.setFont(JvSwingUtils.getLabelFont());
    optimizeOrder.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        orderByAvWidth();
      }
    });

    JButton sortByScore = new JButton(
            MessageManager.getString("label.seq_sort_by_score"));
    sortByScore.setFont(JvSwingUtils.getLabelFont());
    sortByScore.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        af.avc.sortAlignmentByFeatureScore(null);
      }
    });
    JButton sortByDens = new JButton(
            MessageManager.getString("label.sequence_sort_by_density"));
    sortByDens.setFont(JvSwingUtils.getLabelFont());
    sortByDens.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        af.avc.sortAlignmentByFeatureDensity(null);
      }
    });

    JButton help = new JButton(MessageManager.getString("action.help"));
    help.setFont(JvSwingUtils.getLabelFont());
    help.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        try
        {
          Help.showHelpWindow(HelpId.SequenceFeatureSettings);
        } catch (HelpSetException e1)
        {
          e1.printStackTrace();
        }
      }
    });
    help.setFont(JvSwingUtils.getLabelFont());
    help.setText(MessageManager.getString("action.help"));
    help.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        try
        {
          Help.showHelpWindow(HelpId.SequenceFeatureSettings);
        } catch (HelpSetException e1)
        {
          e1.printStackTrace();
        }
      }
    });

    JButton cancel = new JButton(MessageManager.getString("action.cancel"));
    cancel.setFont(JvSwingUtils.getLabelFont());
    cancel.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        fr.setTransparency(originalTransparency);
        fr.setFeatureFilters(originalFilters);
        updateFeatureRenderer(originalData);
        close();
      }
    });

    JButton ok = new JButton(MessageManager.getString("action.ok"));
    ok.setFont(JvSwingUtils.getLabelFont());
    ok.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        close();
      }
    });

    JButton loadColours = new JButton(
            MessageManager.getString("label.load_colours"));
    loadColours.setFont(JvSwingUtils.getLabelFont());
    loadColours.setToolTipText(
            MessageManager.getString("label.load_colours_tooltip"));
    loadColours.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        load();
      }
    });

    JButton saveColours = new JButton(
            MessageManager.getString("label.save_colours"));
    saveColours.setFont(JvSwingUtils.getLabelFont());
    saveColours.setToolTipText(
            MessageManager.getString("label.save_colours_tooltip"));
    saveColours.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        save();
      }
    });
    transparency.addChangeListener(new ChangeListener()
    {
      @Override
      public void stateChanged(ChangeEvent evt)
      {
        if (!inConstruction)
        {
          fr.setTransparency((100 - transparency.getValue()) / 100f);
          af.alignPanel.paintAlignment(true, true);
        }
      }
    });

    summaryView = new JCheckBox(
            MessageManager.getString("label.summary_view"));
    summaryView
            .setToolTipText(
                    MessageManager.getString("label.summary_view_tip"));
    summaryView.addActionListener(new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent e)
      {
        resetTable(null);
      }
    });

    transparency.setMaximum(70);
    transparency.setToolTipText(
            MessageManager.getString("label.transparency_tip"));

    JPanel lowerPanel = new JPanel(new GridLayout(1, 2));
    bigPanel.add(lowerPanel, BorderLayout.SOUTH);

    JPanel transbuttons = new JPanel(new GridLayout(5, 1));
    transbuttons.add(optimizeOrder);
    transbuttons.add(invert);
    transbuttons.add(sortByScore);
    transbuttons.add(sortByDens);
    transbuttons.add(help);
    JPanel transPanel = new JPanel(new GridLayout(3, 1));
    transPanel.add(summaryView);
    transPanel.add(new JLabel(" Colour transparency" + ":"));
    transPanel.add(transparency);
    lowerPanel.add(transPanel);
    lowerPanel.add(transbuttons);

    JPanel buttonPanel = new JPanel();
    buttonPanel.add(ok);
    buttonPanel.add(cancel);
    buttonPanel.add(loadColours);
    buttonPanel.add(saveColours);
    bigPanel.add(scrollPane, BorderLayout.CENTER);
    settingsPane.add(bigPanel, BorderLayout.CENTER);
    settingsPane.add(buttonPanel, BorderLayout.SOUTH);
    this.add(settingsPane);
  }

  /**
   * Reorders features by 'dragging' selectedRow to 'newRow'
   * 
   * @param newRow
   */
  protected void dragRow(int newRow)
  {
    if (summaryView.isSelected())
    {
      // no drag while in summary view
      return;
    }

    if (newRow != selectedRow && selectedRow != -1 && newRow != -1)
    {
      /*
       * reposition 'selectedRow' to 'newRow' (the dragged to location)
       * this could be more than one row away for a very fast drag action
       * so just swap it with adjacent rows until we get it there
       */
      Object[][] data = ((FeatureTableModel) table.getModel())
              .getData();
      int direction = newRow < selectedRow ? -1 : 1;
      for (int i = selectedRow; i != newRow; i += direction)
      {
        Object[] temp = data[i];
        data[i] = data[i + direction];
        data[i + direction] = temp;
      }
      updateFeatureRenderer(data);
      table.repaint();
      selectedRow = newRow;
    }
  }

  protected void refreshTable()
  {
    Object[][] data = ((FeatureTableModel) table.getModel()).getData();
    for (Object[] row : data)
    {
      String type = (String) row[TYPE_COLUMN];
      FeatureColourI colour = fr.getFeatureColours().get(type);
      FeatureMatcherSetI filter = fr.getFeatureFilter(type);
      if (filter == null)
      {
        filter = new FeatureMatcherSet();
      }
      row[COLOUR_COLUMN] = colour;
      row[FILTER_COLUMN] = filter;
    }
    repaint();
  }

  // ///////////////////////////////////////////////////////////////////////
  // http://java.sun.com/docs/books/tutorial/uiswing/components/table.html
  // ///////////////////////////////////////////////////////////////////////
  class FeatureTableModel extends AbstractTableModel
  {
    private String[] columnNames = {
        MessageManager.getString("label.feature_type"),
        MessageManager.getString("action.colour"),
        MessageManager.getString("label.filter"),
        MessageManager.getString("label.show") };

    private Object[][] data;

    FeatureTableModel(Object[][] data)
    {
      this.data = data;
    }

    public Object[][] getData()
    {
      return data;
    }

    public void setData(Object[][] data)
    {
      this.data = data;
    }

    @Override
    public int getColumnCount()
    {
      return columnNames.length;
    }

    public Object[] getRow(int row)
    {
      return data[row];
    }

    @Override
    public int getRowCount()
    {
      return data.length;
    }

    @Override
    public String getColumnName(int col)
    {
      return columnNames[col];
    }

    @Override
    public Object getValueAt(int row, int col)
    {
      return data[row][col];
    }

    /**
     * Answers the class of column c of the table
     */
    @Override
    public Class<?> getColumnClass(int c)
    {
      switch (c)
      {
      case TYPE_COLUMN:
        return String.class;
      case COLOUR_COLUMN:
        return FeatureColour.class;
      case FILTER_COLUMN:
        return FeatureMatcherSet.class;
      default:
        return Boolean.class;
      }
    }

    /**
     * Answers true for all columns except Feature Type
     */
    @Override
    public boolean isCellEditable(int row, int col)
    {
      return col != TYPE_COLUMN;
    }

    /**
     * Sets the value in the model for a given row and column. If Visibility
     * (Show/Hide) is being set, and the table is in Summary View, then it is
     * set also on any sub-types of the row's feature type.
     */
    @Override
    public void setValueAt(Object value, int row, int col)
    {
      data[row][col] = value;
      fireTableCellUpdated(row, col);
      if (summaryView.isSelected() && col == SHOW_COLUMN)
      {
        setSubtypesVisibility(row, (Boolean) value);
      }
      updateFeatureRenderer(data);
    }

    /**
     * Sets the visibility of any feature types which are sub-types of the type
     * in the given row of the table
     * 
     * @param row
     * @param value
     */
    protected void setSubtypesVisibility(int row, Boolean value)
    {
      String type = (String) data[row][TYPE_COLUMN];
      OntologyI so = SequenceOntologyFactory.getInstance();

      for (int r = 0; r < data.length; r++)
      {
        if (r != row)
        {
          String type2 = (String) data[r][TYPE_COLUMN];
          if (so.isA(type2, type))
          {
            data[r][SHOW_COLUMN] = value;
            fireTableCellUpdated(r, SHOW_COLUMN);
          }
        }
      }
    }
  }

  class ColorRenderer extends JLabel implements TableCellRenderer
  {
    javax.swing.border.Border unselectedBorder = null;

    javax.swing.border.Border selectedBorder = null;

    final String baseTT = "Click to edit, right/apple click for menu.";

    public ColorRenderer()
    {
      setOpaque(true); // MUST do this for background to show up.
      setHorizontalTextPosition(SwingConstants.CENTER);
      setVerticalTextPosition(SwingConstants.CENTER);
    }

    @Override
    public Component getTableCellRendererComponent(JTable tbl, Object color,
            boolean isSelected, boolean hasFocus, int row, int column)
    {
      FeatureColourI cellColour = (FeatureColourI) color;
      setOpaque(true);
      setToolTipText(baseTT);
      setBackground(tbl.getBackground());
      if (!cellColour.isSimpleColour())
      {
        Rectangle cr = tbl.getCellRect(row, column, false);
        FeatureSettings.renderGraduatedColor(this, cellColour,
                (int) cr.getWidth(), (int) cr.getHeight());
      }
      else
      {
        this.setText("");
        this.setIcon(null);
        setBackground(cellColour.getColour());
      }
      if (isSelected)
      {
        if (selectedBorder == null)
        {
          selectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
                  tbl.getSelectionBackground());
        }
        setBorder(selectedBorder);
      }
      else
      {
        if (unselectedBorder == null)
        {
          unselectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
                  tbl.getBackground());
        }
        setBorder(unselectedBorder);
      }

      return this;
    }
  }

  class FilterRenderer extends JLabel implements TableCellRenderer
  {
    javax.swing.border.Border unselectedBorder = null;

    javax.swing.border.Border selectedBorder = null;

    public FilterRenderer()
    {
      setOpaque(true); // MUST do this for background to show up.
      setHorizontalTextPosition(SwingConstants.CENTER);
      setVerticalTextPosition(SwingConstants.CENTER);
    }

    @Override
    public Component getTableCellRendererComponent(JTable tbl,
            Object filter, boolean isSelected, boolean hasFocus, int row,
            int column)
    {
      FeatureMatcherSetI theFilter = (FeatureMatcherSetI) filter;
      setOpaque(true);
      String asText = theFilter.toString();
      setBackground(tbl.getBackground());
      this.setText(asText);
      this.setIcon(null);

      if (isSelected)
      {
        if (selectedBorder == null)
        {
          selectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
                  tbl.getSelectionBackground());
        }
        setBorder(selectedBorder);
      }
      else
      {
        if (unselectedBorder == null)
        {
          unselectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
                  tbl.getBackground());
        }
        setBorder(unselectedBorder);
      }

      return this;
    }
  }

  /**
   * update comp using rendering settings from gcol
   * 
   * @param comp
   * @param gcol
   */
  public static void renderGraduatedColor(JLabel comp, FeatureColourI gcol)
  {
    int w = comp.getWidth(), h = comp.getHeight();
    if (w < 20)
    {
      w = (int) comp.getPreferredSize().getWidth();
      h = (int) comp.getPreferredSize().getHeight();
      if (w < 20)
      {
        w = 80;
        h = 12;
      }
    }
    renderGraduatedColor(comp, gcol, w, h);
  }

  public static void renderGraduatedColor(JLabel comp, FeatureColourI gcol,
          int w, int h)
  {
    boolean thr = false;
    StringBuilder tt = new StringBuilder();
    StringBuilder tx = new StringBuilder();

    if (gcol.isColourByAttribute())
    {
      tx.append(String.join(":", gcol.getAttributeName()));
    }
    else if (!gcol.isColourByLabel())
    {
      tx.append(MessageManager.getString("label.score"));
    }
    tx.append(" ");
    if (gcol.isAboveThreshold())
    {
      thr = true;
      tx.append(">");
      tt.append("Thresholded (Above ").append(gcol.getThreshold())
              .append(") ");
    }
    if (gcol.isBelowThreshold())
    {
      thr = true;
      tx.append("<");
      tt.append("Thresholded (Below ").append(gcol.getThreshold())
              .append(") ");
    }
    if (gcol.isColourByLabel())
    {
      tt.append("Coloured by label text. ").append(tt);
      if (thr)
      {
        tx.append(" ");
      }
      if (!gcol.isColourByAttribute())
      {
        tx.append("Label");
      }
      comp.setIcon(null);
    }
    else
    {
      Color newColor = gcol.getMaxColour();
      comp.setBackground(newColor);
      // System.err.println("Width is " + w / 2);
      Icon ficon = new FeatureIcon(gcol, comp.getBackground(), w, h, thr);
      comp.setIcon(ficon);
      // tt+="RGB value: Max (" + newColor.getRed() + ", "
      // + newColor.getGreen() + ", " + newColor.getBlue()
      // + ")\nMin (" + minCol.getRed() + ", " + minCol.getGreen()
      // + ", " + minCol.getBlue() + ")");
    }
    comp.setHorizontalAlignment(SwingConstants.CENTER);
    comp.setText(tx.toString());
    if (tt.length() > 0)
    {
      if (comp.getToolTipText() == null)
      {
        comp.setToolTipText(tt.toString());
      }
      else
      {
        comp.setToolTipText(
                tt.append(" ").append(comp.getToolTipText()).toString());
      }
    }
  }

  class ColorEditor extends AbstractCellEditor
          implements TableCellEditor, ActionListener
  {
    FeatureSettings me;

    FeatureColourI currentColor;

    FeatureTypeSettings chooser;

    String type;

    JButton colourButton;

    JColorChooser colorChooser;

    JDialog dialog;

    protected static final String EDIT = "edit";

    int rowSelected = 0;

    public ColorEditor(FeatureSettings me)
    {
      this.me = me;
      // Set up the editor (from the table's point of view),
      // which is a button.
      // This button brings up the color chooser dialog,
      // which is the editor from the user's point of view.
      colourButton = new JButton();
      colourButton.setActionCommand(EDIT);
      colourButton.addActionListener(this);
      colourButton.setBorderPainted(false);
      // Set up the dialog that the button brings up.
      colorChooser = new JColorChooser();
      dialog = JColorChooser.createDialog(colourButton,
              MessageManager.getString("label.select_colour"), true, // modal
              colorChooser, this, // OK button handler
              null); // no CANCEL button handler
    }

    /**
     * Handles events from the editor button and from the dialog's OK button.
     */
    @Override
    public void actionPerformed(ActionEvent e)
    {
      // todo test e.getSource() instead here
      if (EDIT.equals(e.getActionCommand()))
      {
        // The user has clicked the cell, so
        // bring up the dialog.
        if (currentColor.isSimpleColour())
        {
          // bring up simple color chooser
          colourButton.setBackground(currentColor.getColour());
          colorChooser.setColor(currentColor.getColour());
          dialog.setVisible(true);
        }
        else
        {
          // bring up graduated chooser.
          chooser = new FeatureTypeSettings(me.fr, type);
          chooser.setRequestFocusEnabled(true);
          chooser.requestFocus();
          chooser.addActionListener(this);
          // Make the renderer reappear.
          fireEditingStopped();
        }
      }
      else
      {
        if (currentColor.isSimpleColour())
        {
          /*
           * read off colour picked in colour chooser after OK pressed
           */
          currentColor = new FeatureColour(colorChooser.getColor());
          me.table.setValueAt(currentColor, rowSelected, COLOUR_COLUMN);
        }
        else
        {
          /*
           * after OK in variable colour dialog, any changes to colour 
           * (or filters!) are already set in FeatureRenderer, so just
           * update table data without triggering updateFeatureRenderer
           */
          refreshTable();
        }
        fireEditingStopped();
        me.table.validate();
      }
    }

    // Implement the one CellEditor method that AbstractCellEditor doesn't.
    @Override
    public Object getCellEditorValue()
    {
      return currentColor;
    }

    // Implement the one method defined by TableCellEditor.
    @Override
    public Component getTableCellEditorComponent(JTable theTable, Object value,
            boolean isSelected, int row, int column)
    {
      currentColor = (FeatureColourI) value;
      this.rowSelected = row;
      type = me.table.getValueAt(row, TYPE_COLUMN).toString();
      colourButton.setOpaque(true);
      colourButton.setBackground(me.getBackground());
      if (!currentColor.isSimpleColour())
      {
        JLabel btn = new JLabel();
        btn.setSize(colourButton.getSize());
        FeatureSettings.renderGraduatedColor(btn, currentColor);
        colourButton.setBackground(btn.getBackground());
        colourButton.setIcon(btn.getIcon());
        colourButton.setText(btn.getText());
      }
      else
      {
        colourButton.setText("");
        colourButton.setIcon(null);
        colourButton.setBackground(currentColor.getColour());
      }
      return colourButton;
    }
  }

  /**
   * The cell editor for the Filter column. It displays the text of any filters
   * for the feature type in that row (in full as a tooltip, possible abbreviated
   * as display text). On click in the cell, opens the Feature Display Settings
   * dialog at the Filters tab.
   */
  class FilterEditor extends AbstractCellEditor
          implements TableCellEditor, ActionListener
  {
    FeatureSettings me;

    FeatureMatcherSetI currentFilter;

    Point lastLocation;

    String type;

    JButton filterButton;

    protected static final String EDIT = "edit";

    int rowSelected = 0;

    public FilterEditor(FeatureSettings me)
    {
      this.me = me;
      filterButton = new JButton();
      filterButton.setActionCommand(EDIT);
      filterButton.addActionListener(this);
      filterButton.setBorderPainted(false);
    }

    /**
     * Handles events from the editor button
     */
    @Override
    public void actionPerformed(ActionEvent e)
    {
      if (filterButton == e.getSource())
      {
        FeatureTypeSettings chooser = new FeatureTypeSettings(me.fr, type);
        chooser.addActionListener(this);
        chooser.setRequestFocusEnabled(true);
        chooser.requestFocus();
        if (lastLocation != null)
        {
          // todo open at its last position on screen
          chooser.setBounds(lastLocation.x, lastLocation.y,
                  chooser.getWidth(), chooser.getHeight());
          chooser.validate();
        }
        fireEditingStopped();
      }
      else if (e.getSource() instanceof Component)
      {
        /*
         * after OK in variable colour dialog, any changes to filter
         * (or colours!) are already set in FeatureRenderer, so just
         * update table data without triggering updateFeatureRenderer
         */
        refreshTable();
        fireEditingStopped();
        me.table.validate();
      }
    }

    @Override
    public Object getCellEditorValue()
    {
      return currentFilter;
    }

    @Override
    public Component getTableCellEditorComponent(JTable theTable, Object value,
            boolean isSelected, int row, int column)
    {
      currentFilter = (FeatureMatcherSetI) value;
      this.rowSelected = row;
      type = me.table.getValueAt(row, TYPE_COLUMN).toString();
      filterButton.setOpaque(true);
      filterButton.setBackground(me.getBackground());
      filterButton.setText(currentFilter.toString());
      filterButton.setToolTipText(currentFilter.toString());
      filterButton.setIcon(null);
      return filterButton;
    }
  }
}

class FeatureIcon implements Icon
{
  private static final Font VERDANA_9 = new Font("Verdana", Font.PLAIN, 9);

  FeatureColourI gcol;

  Color backg;

  boolean midspace = false;

  int width = 50, height = 20;

  int s1, e1; // start and end of midpoint band for thresholded symbol

  Color mpcolour = Color.white;

  FeatureIcon(FeatureColourI gfc, Color bg, int w, int h, boolean mspace)
  {
    gcol = gfc;
    backg = bg;
    width = w;
    height = h;
    midspace = mspace;
    if (midspace)
    {
      s1 = width / 3;
      e1 = s1 * 2;
    }
    else
    {
      s1 = width / 2;
      e1 = s1;
    }
  }

  @Override
  public int getIconWidth()
  {
    return width;
  }

  @Override
  public int getIconHeight()
  {
    return height;
  }

  @Override
  public void paintIcon(Component c, Graphics g, int x, int y)
  {

    if (gcol.isColourByLabel())
    {
      g.setColor(backg);
      g.fillRect(0, 0, width, height);
      // need an icon here.
      g.setColor(gcol.getMaxColour());

      g.setFont(VERDANA_9);

      // g.setFont(g.getFont().deriveFont(
      // AffineTransform.getScaleInstance(
      // width/g.getFontMetrics().stringWidth("Label"),
      // height/g.getFontMetrics().getHeight())));

      g.drawString(MessageManager.getString("label.label"), 0, 0);

    }
    else
    {
      Color minCol = gcol.getMinColour();
      g.setColor(minCol);
      g.fillRect(0, 0, s1, height);
      if (midspace)
      {
        g.setColor(Color.white);
        g.fillRect(s1, 0, e1 - s1, height);
      }
      g.setColor(gcol.getMaxColour());
      g.fillRect(0, e1, width - e1, height);
    }
  }
}
